from collections import abc
from copy import deepcopy
from dataclasses import asdict, is_dataclass
from functools import reduce
from inspect import isclass
from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple, TypeVar, Union
from typing import cast as typing_cast
from typing import overload

from omegaconf import DictConfig, ListConfig, OmegaConf
from omegaconf.errors import MissingMandatoryValue
from omegaconf.omegaconf import DictKeyType

from .flexyclasses import FlexyClass
from .traversal import FailedParamSpec, traverse

C = TypeVar("C", bound=Union[DictConfig, ListConfig])


def cast(config: Any, copy: bool = False) -> DictConfig:
    if is_dataclass(config):
        parsed_config = from_dataclass(config)
    elif isinstance(config, dict):
        parsed_config = from_dict(config)
    elif isinstance(config, str):
        parsed_config = from_string(config)
    elif config is None:
        parsed_config = from_none(config)
    elif isinstance(config, DictConfig):
        parsed_config = deepcopy(config) if copy else config
    elif isinstance(config, Path):
        parsed_config = from_file(config)
    else:
        raise TypeError(f"Cannot cast `{type(config)}` to DictConfig")

    return parsed_config


def from_none(*args: Any, **kwargs: Any) -> DictConfig:
    """Returns an empty dict config"""
    return OmegaConf.create()


def from_dataclass(config: Any) -> DictConfig:
    """Cast a dataclass to a structured omega config"""
    if config is None:
        return from_none()

    if isclass(config) and issubclass(config, FlexyClass):
        config = config.defaults

    elif not is_dataclass(config):
        raise TypeError(f"`{config}` is not a dataclass!")

    parsed_config = OmegaConf.structured(config)

    if not isinstance(parsed_config, DictConfig):
        raise TypeError(f"Cannot create dict config from `{config}`")
    return parsed_config


@overload
def from_python(
    config: Union[Dict[DictKeyType, Any], Dict[str, Any]]
) -> DictConfig:
    ...


@overload
def from_python(config: List[Any]) -> ListConfig:
    ...


def from_python(
    config: Union[Dict[DictKeyType, Any], Dict[str, Any], List[Any]]
) -> Union[DictConfig, ListConfig]:
    """Create a config from a dict"""
    if not isinstance(config, (dict, list)):
        raise TypeError(f"`{config}` is not a dict or list!")

    parsed_config = OmegaConf.create(config)

    if not isinstance(parsed_config, (DictConfig, ListConfig)):
        raise ValueError(
            f"Config `{config}` is not a DictConfig or ListConfig!"
        )

    return parsed_config


def from_dict(
    config: Union[Dict[DictKeyType, Any], Dict[str, Any]]
) -> DictConfig:
    """Create a config from a dict"""
    if not isinstance(config, dict):
        raise TypeError(f"`{config}` is not a dict!")

    return from_python(config)  # type: ignore


def from_string(config: str) -> DictConfig:
    """Load a config from a string"""
    if not isinstance(config, str):
        raise TypeError(f"`{config}` is not a string!")

    parsed_config = OmegaConf.create(config)

    if not isinstance(parsed_config, DictConfig):
        raise ValueError(f"Config `{config}` is not a DictConfig!")
    return parsed_config


def from_file(path: Union[str, Path]) -> DictConfig:
    """Load a config from a file"""
    path = Path(path)

    if not path.exists():
        raise FileNotFoundError(f"Cannot file configuration at {path}")
    config = OmegaConf.load(path)

    if not isinstance(config, DictConfig):
        raise ValueError(f"Config loaded from {path} is not a DictConfig!")

    return config


def from_options(opts: Sequence[str]) -> DictConfig:
    """Create a config from a list of options"""
    if not isinstance(opts, abc.Sequence) or not all(
        isinstance(o, str) for o in opts
    ):
        raise TypeError(f"`{opts}` is not a list of strings!")

    config = OmegaConf.from_dotlist(list(opts))
    if not isinstance(config, DictConfig):
        raise TypeError(f"input is not a sequence of strings, but `{opts}")
    return config


def to_yaml(config: Any) -> str:
    """Convert a omegaconf config to a YAML string"""
    if not isinstance(config, DictConfig):
        config = from_dataclass(config)
    return OmegaConf.to_yaml(config)


def to_python(config: Any) -> Any:
    """Convert a omegaconf config to a Python primitive type"""
    if is_dataclass(config):
        config = from_dataclass(config)
    container = OmegaConf.to_container(config)
    return container


def to_dict(config: Any) -> Dict[DictKeyType, Any]:
    """Convert a omegaconf config to a Python primitive type"""
    container = to_python(config)

    if not isinstance(container, dict):
        raise TypeError(f"`{container}` is not a dict!")

    return container


########################################


def resolve(config: C) -> C:
    """Resolve a config"""
    config = deepcopy(config)
    OmegaConf.resolve(config)
    return config


def safe_validate(config_node: C) -> Tuple[C, List[FailedParamSpec]]:
    """Check if all attributes are resolve and not missing

    If resolution fails, second element of return tuple contains
    missing keys."""

    errors = []

    if not isinstance(config_node, (DictConfig, ListConfig)):
        raise TypeError(f"`{config_node}` is not a DictConfig!")

    for spec in traverse(config_node):
        if spec.key is None:
            raise RuntimeError(
                "You should not be here! Something went "
                "wrong in the core Springs library."
            )

        if OmegaConf.is_missing(spec.parent, spec.key):
            # raise ValueError(f"Missing value for `{spec.path}`")
            errors.append(
                FailedParamSpec(error=MissingMandatoryValue(), **asdict(spec))
            )

        if OmegaConf.is_interpolation(spec.parent, spec.key):
            try:
                getattr(spec.parent, str(spec.key))
            except Exception as e:
                errors.append(FailedParamSpec(error=e, **asdict(spec)))

    return config_node, errors


def validate(
    config_node: C,
    raise_on_missing: bool = True,
    raise_on_failed_interpolation: bool = True,
) -> C:
    config_node, errors = safe_validate(config_node)

    for spec in errors:
        if spec.interpol and not raise_on_failed_interpolation:
            continue
        if (
            isinstance(spec.error, MissingMandatoryValue)
            and not raise_on_missing
        ):
            continue

        # this will not run if missing is an empty list
        raise ValueError(
            f"Interpolation for '{spec.path}' not resolved; "
            f"{type(spec.error).__name__}: "
            f"{' '.join(str(a) for a in spec.error.args)}"
        ) from spec.error
    return config_node


def unsafe_merge(
    first_config: C,
    *other_configs: Union[DictConfig, ListConfig],
) -> C:
    """Merge two or more configurations together without any validation.

    Args:
        first_config (Union[DictConfig, ListConfig]): The first configuration
            to merge.
        *other_configs (Union[DictConfig, ListConfig]): The other
            configurations to merge with the first one; they are merged in
            the order they are provided.

    Returns:
        Union[DictConfig, ListConfig]: The merged configuration.
    """
    for c in (first_config, *other_configs):
        if not isinstance(c, (DictConfig, ListConfig)):
            raise TypeError(f"`{c}` is not a DictConfig or ListConfig!")

    output_config = reduce(
        lambda a, b: OmegaConf.merge(a, OmegaConf.create(b)),
        (
            OmegaConf.to_container(c, resolve=False)
            for c in (first_config, *other_configs)
        ),
        # we need to start with an empty dict or list rather than
        # the first config to prevent type checking from kicking in
        # during merge.
        OmegaConf.create({} if isinstance(first_config, DictConfig) else []),
    )

    return typing_cast(C, output_config)


def merge(
    first_config: C,
    *other_configs: Union[DictConfig, ListConfig],
    resolve_nodes: bool = True,
    validate_config: bool = True,
) -> C:
    """Merge two or more configurations together. If `skip_resolve`
    and `skip_validate` are both `True`, this function is equivalent to
    `unsafe_merge`.

    Args:
        first_config (Union[DictConfig, ListConfig]): The first configuration
            to merge.
        *other_configs (Union[DictConfig, ListConfig]): The other
            configurations to merge with the first one; they are merged in
            the order they are provided.
        resolve_nodes (bool, optional): If `False`, keys that are references
            to other keys will not be resolved. Defaults to `True`.
        validate_config (bool, optional): If `False`, the merged configuration
            will not be validated. Defaults to `True`.

    Returns:
        Union[DictConfig, ListConfig]: The merged configuration.
    """

    unsafe_config = unsafe_merge(first_config, *other_configs)

    if resolve_nodes:
        unsafe_config = resolve(unsafe_config)

    output_config = OmegaConf.merge(first_config, unsafe_config)

    if validate_config:
        output_config = validate(
            output_config,
            raise_on_failed_interpolation=resolve_nodes,
        )

    return typing_cast(C, output_config)


L = TypeVar("L", bound=ListConfig)


def edit_list(config: L, editor: DictConfig) -> L:
    """Edit a copy of a ListConfig using a DictConfig.

    Args:
        config (ListConfig): The ListConfig to edit.
        editor (DictConfig): The DictConfig containing the edits to make.
            The keys of the DictConfig should be the indices of the list.

    Returns:
        ListConfig: The edited ListConfig.
    """
    if not isinstance(config, ListConfig):
        raise TypeError(f"`{config}` is not a ListConfig!")

    if not isinstance(editor, DictConfig):
        raise TypeError(f"`{editor}` is not a DictConfig!")

    config = deepcopy(config)

    for key, value in editor.items():
        try:
            key = int(key)  # pyright: reportGeneralTypeIssues=false
        except TypeError:
            raise TypeError(f"`{key}` is not an int!")

        if key >= len(config):
            raise IndexError(f"Index {key} is out of range!")

        if isinstance(config[key], (DictConfig, ListConfig)):
            config[key] = merge(config[key], value)
        else:
            config[key] = value

    return config



from copy import deepcopy
from dataclasses import MISSING as DT_MISSING
from dataclasses import Field, field, fields, is_dataclass
from typing import Any, Dict, Generic, Type, TypeVar

from omegaconf import MISSING as OC_MISSING
from typing_extensions import dataclass_transform

from .utils import get_annotations

C = TypeVar("C", bound=Any)


class FlexyClass(dict, Generic[C]):
    """A FlexyClass is a dictionary with some default values assigned to it
    FlexyClasses are generally not used directly, but rather creating using
    the `flexyclass` decorator.

    NOTE: When instantiating a new FlexyClass object directly, the constructor
    actually returns a `dataclasses.Field` object. This is for API consistency
    with how dataclasses are used in a structured configuration. If you want to
    access values in the FlexyClass directly, use FlexyClass.defaults property.
    """

    __origin__: type = dict
    __flexyclass_defaults__: Dict[str, Any] = {}

    @classmethod
    def _unpack_if_dataclass_field(cls, value: Any) -> Any:
        if isinstance(value, Field):
            if value.default_factory is not DT_MISSING:
                value = value.default_factory()
            elif value.default is not DT_MISSING:
                value = value.default
            else:
                value = OC_MISSING
        else:
            value = value

        if isinstance(value, dict):
            value = {
                k: cls._unpack_if_dataclass_field(v) for k, v in value.items()
            }
        elif isinstance(value, list):
            value = [cls._unpack_if_dataclass_field(v) for v in value]

        return value

    @classmethod
    @property
    def defaults(cls):
        """The default values for the FlexyClass"""

        return cls._unpack_if_dataclass_field(
            deepcopy(cls.__flexyclass_defaults__)
        )

    def __new__(cls, **kwargs):
        # We completely change how the constructor works to allow users
        # to use flexyclasses in the same way they would use a dataclass.
        factory_dict: Dict[str, Any] = {}
        factory_dict = {**cls.defaults, **kwargs}
        return field(default_factory=lambda: factory_dict)

    @classmethod
    def flexyclass(cls, target_cls: Type[C]) -> Type["FlexyClass"]:
        """Decorator to create a FlexyClass from a class"""

        if is_dataclass(target_cls):
            attributes_iterator = ((f.name, f) for f in fields(target_cls))
        else:
            attributes_iterator = (
                (f_name, getattr(target_cls, f_name, OC_MISSING))
                for f_name in get_annotations(target_cls)
            )

        defaults = {
            f_name: cls._unpack_if_dataclass_field(f_value)
            for f_name, f_value in attributes_iterator
        }

        return type(
            target_cls.__name__,
            (FlexyClass,),
            {"__flexyclass_defaults__": defaults},
        )


@dataclass_transform()
def flexyclass(cls: Type[C]) -> Type[FlexyClass[C]]:
    """Alias for FlexyClass.flexyclass"""
    return FlexyClass.flexyclass(cls)



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar

from omegaconf import MISSING, DictConfig, ListConfig

from .commandline import cli
from .core import (
    cast,
    edit_list,
    from_dataclass,
    from_dict,
    from_file,
    from_none,
    from_options,
    from_string,
    merge,
    resolve,
    to_dict,
    to_yaml,
    unsafe_merge,
    validate,
)
from .flexyclasses import flexyclass
from .initialize import Target, init
from .logging import configure_logging
from .nicknames import NicknameRegistry
from .resolvers import all_resolvers, register
from .traversal import traverse
from .types import get_type
from .utils import SpringsWarnings, get_version

__version__ = get_version()


T = TypeVar("T")


def toggle_warnings(value: Optional[bool] = None):
    """Shortcut for springs.utils.SpringsWarnings.toggle"""
    SpringsWarnings.toggle(value)


def make_target(c: Callable) -> str:
    """Shortcut for springs.initialize.Target.to_string"""
    return Target.to_string(c)


def nickname(name: str) -> Callable[[Type[T]], Type[T]]:
    """Shortcut for springs.nicknames.NicknameRegistry.add"""
    return NicknameRegistry.add(name)


def make_flexy(cls_: Any) -> Any:
    """Shortcut for springs.flexyclasses.flexyclass"""

    SpringsWarnings.deprecated(
        deprecated="make_flexy",
        replacement="flexyclass",
    )
    return flexyclass(cls_)


def fdict(**kwargs: Any) -> Dict[str, Any]:
    """Shortcut for creating a Field with a default_factory that returns
    a dictionary.

    Args:
        **kwargs: values for the dictionary returned by default factory"""

    def _factory_fn() -> Dict[str, Any]:
        return {**kwargs}

    return field(default_factory=_factory_fn)


def flist(*args: Any) -> List[Any]:
    """Shortcut for creating a Field with a default_factory that returns
    a list.

    Args:
        *args: values for the list returned by default factory"""

    def _factory_fn() -> List[Any]:
        return [*args]

    return field(default_factory=_factory_fn)


__all__ = [
    "all_resolvers",
    "cast",
    "cli",
    "configure_logging",
    "dataclass",
    "DictConfig",
    "edit_list",
    "fdict",
    "field",
    "flexyclass",
    "flist",
    "from_dataclass",
    "from_dict",
    "from_file",
    "from_none",
    "from_options",
    "from_string",
    "get_type",
    "init",
    "ListConfig",
    "make_flexy",
    "make_target",
    "merge",
    "MISSING",
    "nickname",
    "register",
    "resolve",
    "Target",
    "to_dict",
    "to_yaml",
    "toggle_warnings",
    "traverse",
    "unsafe_merge",
    "validate",
]



from dataclasses import dataclass
from typing import Any, Iterator, Optional, Union
from typing import cast as typecast

from omegaconf import MISSING, DictConfig, ListConfig, OmegaConf
from omegaconf.basecontainer import BaseContainer


@dataclass
class ParamSpec:
    key: Union[str, int, None]
    path: str
    value: Any
    parent: Optional[Union[DictConfig, ListConfig]]
    interpol: bool

    def is_node(self) -> bool:
        return isinstance(self.value, (DictConfig, ListConfig))


@dataclass
class FailedParamSpec(ParamSpec):
    error: Exception


def traverse(
    node: BaseContainer,
    include_nodes: bool = False,
    include_leaves: bool = True,
    include_root: bool = False,
    recurse: bool = True,
) -> Iterator[ParamSpec]:

    if include_root:
        yield ParamSpec(
            key=None, path="", value=node, parent=None, interpol=False
        )

    if isinstance(node, ListConfig):
        for key in range(len(node)):
            if is_interpolation := OmegaConf.is_interpolation(node, key):
                # we don't want to resolve interpolations for now
                # as they might not be resolvable at the moment
                value = typecast(list, OmegaConf.to_container(node))[
                    typecast(int, key)
                ]
            else:
                value = node[typecast(int, key)]

            if isinstance(value, BaseContainer):
                # we recurse into the node if it is a container
                if recurse:
                    for spec in traverse(
                        value,
                        include_leaves=include_leaves,
                        include_nodes=include_nodes,
                    ):
                        yield ParamSpec(
                            key=spec.key,
                            path=f"[{key}].{spec.path}",
                            value=spec.value,
                            parent=spec.parent,
                            interpol=is_interpolation,
                        )

            if include_nodes or include_leaves:
                # we yield the node if it is not a container, or if
                # we are specifically asked to include nodes
                yield ParamSpec(
                    key=key,
                    # [num] is the notation for list indices
                    path=f"[{key}]",
                    value=value,
                    parent=node,
                    interpol=is_interpolation,
                )

    elif isinstance(node, DictConfig):
        for key in node.keys():
            if OmegaConf.is_missing(node, key):
                value = MISSING
                is_interpolation = False
            elif is_interpolation := OmegaConf.is_interpolation(
                node, str(key)
            ):
                # We don't want to resolve interpolations for now,
                # as they might not be resolvable at the moment
                value = OmegaConf.to_container(node)[key]  # type: ignore
            else:
                value = OmegaConf.select(node, str(key))

            if isinstance(value, BaseContainer):
                if recurse:
                    for spec in traverse(
                        value,
                        include_nodes=include_nodes,
                        include_leaves=include_leaves,
                    ):
                        yield ParamSpec(
                            key=spec.key,
                            # we use str(key) here to make sure
                            # it's just a string, not an obj repr
                            path=f"{str(key)}.{spec.path}",
                            value=spec.value,
                            parent=spec.parent,
                            interpol=is_interpolation,
                        )

            if include_nodes or include_leaves:
                # we yield the node if it is not a container, or if
                # we are specifically asked to include nodes
                yield ParamSpec(
                    key=str(key),
                    path=str(key),
                    value=value,
                    parent=node,
                    interpol=is_interpolation,
                )

    else:
        raise ValueError(
            f"Cannot traverse `{node}`; DictConfig or ListConfig "
            f"expected, but got `{type(node)}`."
        )



import importlib.metadata
import inspect
import os
import re
import shutil
import sys
import warnings
from ast import literal_eval
from pathlib import Path
from textwrap import dedent
from typing import Any, Callable, Dict, Optional, Type, Union

import yaml
from omegaconf import DictConfig, OmegaConf


def clean_multiline(string: str) -> str:
    string = dedent(string.strip())
    string = re.sub(r"\s*\n\s*", " ", string)
    return string


def get_annotations(obj: Any) -> Dict[str, type]:
    # use inspect.get_annotations if python >= 3.10
    if sys.version_info >= (3, 10):
        return inspect.get_annotations(obj)
    else:
        return getattr(obj, "__annotations__", {})


def get_version() -> str:
    """Get the version of the package."""

    # This is a workaround for the fact that if the package is installed
    # in editable mode, the version is not reliability available.
    # Therefore, we check for the existence of a file called EDITABLE,
    # which is not included in the package at distribution time.
    path = Path(__file__).parent / "EDITABLE"
    if path.exists():
        return "dev"

    try:
        # package has been installed, so it has a version number
        # from pyproject.toml
        version = importlib.metadata.version(__package__ or __name__)
    except importlib.metadata.PackageNotFoundError:
        # package hasn't been installed, so set version to "dev"
        version = "dev"

    return version


class PrintUtils:
    """A few utilities to make printing easier"""

    def __init__(
        self,
        indent_step: int = 2,
        indent_char: str = " ",
        separator_char: str = "-",
        terminal_width: Optional[int] = None,
        print_fn: Optional[Callable[..., None]] = None,
    ) -> None:
        self.indent_step = indent_step
        self.indent_char = indent_char
        self.separator_char = separator_char
        self.terminal_width = (
            terminal_width or shutil.get_terminal_size().columns
        )
        self.print_fn = print_fn or print
        self.has_printed_first_separator = False
        self.is_default_print = self.print_fn is print

    def print(
        self,
        *lines: Union[str, dict, DictConfig],
        level: int = 0,
        level_up: Optional[int] = None,
    ):

        level_up = level_up or float("inf")  # type: ignore

        content = ""
        if self.is_default_print:
            if not self.has_printed_first_separator:
                content += self.separator()
            self.has_printed_first_separator = True

        for line in lines:
            if not line:
                # backtrack the new line you just added from previous
                # argument if the current argument evaluates to none, like
                # an empty string, or an empty dict
                content = f"{content.rstrip()} "

            if isinstance(line, DictConfig):
                # we ignore type bc we are ok with line being any of
                # the types returned by omegaconf
                line = OmegaConf.to_container(line)  # type: ignore

            if isinstance(line, dict):
                content += f"{self.to_yaml(line, level)}\n"
            else:
                content += f"{self.indent(str(line), level)}\n"

            level += 1 if level < level_up else 0  # type: ignore

        if self.is_default_print:
            content += f"{self.separator()}\n"

        self.print_fn(content.rstrip())  # type: ignore

    def indent(self, line: str, level: int) -> str:
        indent = self.indent_char * level * self.indent_step
        return indent + line

    def separator(self, level: int = 0) -> str:
        separator = self.separator_char * self.terminal_width
        return self.indent(separator, level=level)

    def to_yaml(self, content: dict, level: int = 0) -> str:
        if not content:
            # this is in case the dict/list is empty
            return str(content)

        out = yaml.dump(
            content,
            indent=self.indent_step,
            width=self.terminal_width,
            default_flow_style=False,
        )
        out = "\n".join(
            self.indent(ln, level) for ln in out.strip().split("\n")
        ).rstrip()
        return out


class SpringsWarnings:
    _WARNINGS = literal_eval(os.environ.get("SPRINGS_WARNINGS", "True"))

    @classmethod
    def toggle(cls, value: Optional[bool] = None):
        if value is None:
            value = not cls._WARNINGS
        cls._WARNINGS = value

    @classmethod
    def _warn(
        cls: Type["SpringsWarnings"],
        message: str,
        category: Type[Warning],
        stacklevel: int = 2,
    ):
        if cls._WARNINGS:
            warnings.warn(message, category, stacklevel=stacklevel)

    @classmethod
    def missing_type(cls: Type["SpringsWarnings"], fn_name: str, type_: Any):
        if type_ is None:
            cls._warn(
                clean_multiline(
                    f"""It is strongly recommended to provide a _type_ argument
                        to `{fn_name}`. This ensures that the correct type is
                        annotated as the return value. Further, it performs
                        type checking on the initialized object."""
                ),
                category=UserWarning,
            )

    @classmethod
    def deprecated(
        cls: Type["SpringsWarnings"],
        deprecated: str,
        removed_when: Optional[str] = None,
        replacement: Optional[str] = None,
    ):
        msg = f"`{deprecated}` is deprecated"
        if removed_when:
            msg += f" and will be removed in Springs {removed_when}."
        else:
            msg += " and may be removed in a future version."

        if replacement:
            msg += f" Use `{replacement}` instead."

        cls._warn(message=msg, category=DeprecationWarning)

    @classmethod
    def argument(cls: Type["SpringsWarnings"], arg_name: str, obj_name: str):
        cls._warn(
            f"'{arg_name}' was provided to `{obj_name}`, but it is ignored",
            category=SyntaxWarning,
        )
